/* 
   nvrec_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>

#include "NVrec.h"

#define ABUFFERSIZE 65536

int nvcore_fin = 0;
int nvcore_die = 0;

void * nvrec_core_thread(void *t)
{
	double f = *(double *)t;
	struct sched_param sp;

	sp.sched_priority = sched_get_priority_min(SCHED_OTHER);
	pthread_setschedparam(pthread_self(), SCHED_OTHER, &sp);

	while (!nvcore_die) {
		sync();
		usleep(1000000.0 * f);
//		fprintf(stderr, ".\n");
	}
	return NULL;
}

void nvcore_quit_handler(int i)
{
	nvcore_fin = 1;
	fprintf(stderr, "nvcore: exiting!\n");
	fprintf(stderr, "  *** If the movier is long, this could take a while, and some oss errors\n");
	fprintf(stderr, "  *** may be displayed.  DO NOT PRESS CTRL-C AGAIN!\n");
}

struct s2 {
	unsigned short lo;
	unsigned short hi;
};

typedef union {
	unsigned int ui;
	struct s2 us;
} ws_t;

int nvcore_main(char **args, struct cap_core_t * cct, struct aud_core_t * audt, struct file_core_t * fct, ntime_t tofs)
{
	ntime_t atime, vtime, fatime = 0, fvtime = 0;
	double capstat, ftemp = 0, lastoff = 0;
	int i = 0, drate = 0, cseq = 0;
	unsigned char abuffer[ABUFFERSIZE];
	unsigned char **vframe;
	unsigned long long fdrop = 0;
	signed long long fcount = -1;
	ws_t dseq;
	pthread_t sthread;
	sigset_t set;
	double Stime = 0, Qlow = 10.0, Qfast = 20.0, Qhi = 40.0, Qscale;
	struct sigaction newaction, oldaction;
	NV_ARGS;
	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-L") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			Qlow = atof(NV_THIS_ARG);
			if((Qlow < 0) || (Qlow > 100))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-M") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			Qfast = atof(NV_THIS_ARG);
			if((Qfast < 0) || (Qfast > 100))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-H") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			Qhi = atof(NV_THIS_ARG);
			if((Qhi < 0) || (Qhi > 100))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-S") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			Stime = atof(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-F") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fcount = atoll(NV_THIS_ARG);
			if(fcount <= 0)
				goto usage;
		}
	}
	
	Qscale = 65536.0 / (100.0 - Qhi);
	sync();
	if(Stime > 0) {
		fprintf(stderr, "nvrec core - using separate sync() thread\n");
		if(pthread_create(&sthread, NULL, nvrec_core_thread, (void *)&Stime) != 0)
			fprintf(stderr, "nvrec core - could not start sync thread: continuing anyway.\n");
	} else if(Stime < 0) {
		fprintf(stderr, "nvrec core - using synchronous writes\n");
	} else {
		fprintf(stderr, "nvrec core - not using any sync() method\n");
	}

	if (sigfillset(&set) != 0) {
		perror("NVrec: could not init signal set");
		return -1;
	}
	newaction.sa_sigaction = NULL;
	newaction.sa_mask = set;
	newaction.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	newaction.sa_restorer = NULL;
	newaction.sa_handler = nvcore_quit_handler;
	if(sigaction(SIGINT, &newaction, &oldaction) < 0) {
		perror("NVrec: could not setsignal handler");
		return -1;
	}
	
	dseq.ui = 0;
	
	/* Get initial synch */
	atime = audt->aud_core_tailtime(audt);
	cct->cap_core_next(cct, &vtime);
	vtime -= tofs;
	if(atime > vtime) {
		do {
			cct->cap_core_next(cct, &vtime);
			vtime -= tofs;
		}  while (atime > vtime);
	} else {
		do {
			audt->aud_core_droptail(audt);
			atime = audt->aud_core_tailtime(audt);
		}  while (atime < vtime);
	}
	if(tofs > 0) {
		vtime += tofs;
		while(atime < vtime) {
			i = audt->aud_core_read(audt, vtime, 0, abuffer, ABUFFERSIZE);
			fct->file_core_audio(fct, abuffer, i);
			atime = audt->aud_core_tailtime(audt);
		}
	}
	fprintf(stderr, "NVrec: initial sync - audio: %lli video %lli (delta %f)\n", atime, vtime, (atime-vtime)/1e9);

	/* Now loop until we're done! */
	while((!nvcore_fin) && (fcount)) {
		if(!(vframe = cct->cap_core_next(cct, &vtime))) {
			fprintf(stderr, "NVrec: error getting more video data. Trying to continue.\n");
			continue;
		}
		capstat = cct->cap_core_status(cct);
		if(capstat > Qhi) {
			dseq.ui += (int)((capstat - Qhi) * Qscale);
		} else if(capstat < Qlow) {
			usleep(cct->cap_core_ftime(cct)/1000);
		}
		if(dseq.us.hi) {
			dseq.us.hi = 0;
			fct->file_core_nullvideo(fct);
			fdrop++;
		} else {
			if(capstat > Qfast)
				fct->file_core_fastvideo(fct, vframe);
			else
				fct->file_core_video(fct, vframe);
		}
		i = audt->aud_core_read(audt, vtime, drate, abuffer, ABUFFERSIZE);
		fct->file_core_audio(fct, abuffer, i);
		
		if(Stime < 0)
			fct->file_core_sync(fct);

		fvtime += cct->cap_core_ftime(cct);
		fatime += i;
		
		atime = audt->aud_core_tailtime(audt);

		if ((cseq % 30) == 0) {
			ftemp = (fatime * 1000000000LL) / audt->aud_core_bps(audt);
			ftemp -= (double)fvtime;
			drate = (int)(- ftemp / 1e6);
			drate -= (int)(lastoff / 1e8);
			lastoff = lastoff + ftemp;
			if(drate >= 1000) drate = 999;
			if(drate <= -1000) drate = -999;
		}
		fprintf(stderr, "VQ:% 8.03f Offset: % 8.03f Slipage: % 8.03f Warp: %d Drop: %lld              \r", capstat, (atime-vtime)/1e9, ftemp / 1e9, drate, fdrop);
		if(fcount > 0) fcount--;
		cseq ++;
	}
	nvcore_die = 1;
	sigaction(SIGINT, &oldaction, NULL);
	return 0;

usage:
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	return -1;
}

void nvcore_usage()
{
	fprintf(stderr, "NVrec core usage:\n");
	fprintf(stderr, "\t-H <%% buffer usage before dropping>  [40]\n");
	fprintf(stderr, "\t-L <min %% buffer usage>              [10]\n");
	fprintf(stderr, "\t-M <%% buffer usage before fast comp> [20]\n");
	fprintf(stderr, "\t-F <number of frames to grab>        [inf]\n");
	fprintf(stderr, "\t-S <seconds between syncs>           [1.0]\n");
}

/*
	Parse the arguments. Define defaults, and set overrides.
	All modules parse arguments left to right, so we set the defaults
	first, then the user arguments, then the overrides, then the filename.
*/

char *nvcore_defaults[] = {"-d", "/dev/dsp",
			   "-r", "44100",
			   "-b", "16",
			   "-v", "/dev/video",
			   "-w", "384",
			   "-h", "288",
			   "-N", "60",
			   "-L", "10",
			   "-H", "40",
			   "-M", "20",
			   "-m",
			   NULL};

char **nvcore_args(int argc, char **argv, char **override)
{
	int i = 0;
	char ** args;
	char ** arg = NULL;
	char ** arg_d = NULL;
	
	for(arg = override; (arg && *arg); arg++) i++;
	for(arg = nvcore_defaults; *arg; arg++) i++;
	args = (char **)malloc(sizeof(char *) * (argc + i + 1));
	if(!args) {
		perror("nvcore args: could not allocate memory!");
		return NULL;
	}
	arg_d = args;
	*(arg_d++) = argv[0];
	for(arg = nvcore_defaults; *arg; arg++)
		*(arg_d++) = *arg;
	for(i = 1; i < argc; i++)
		*(arg_d++) = argv[i];
	for(arg = override; (arg && *arg); arg++)
		*(arg_d++) = *arg;
	*arg_d = NULL;
	return args;
}

void nvcore_close(char ** args)
{
	free(args);
}
